from easyecs.cloudformation.template.task_definition import create_task_definition


def create_template(
    service_name,
    aws_account_id,
    aws_region,
    vpc_id,
    subnet_ids,
    azs,
    ecs_manifest,
    run=False,
):
    from aws_cdk import App, BootstraplessSynthesizer, Environment, Stack
    from aws_cdk.aws_ec2 import Subnet, SubnetSelection, Vpc

    app = App(outdir="./.cloudformation")
    bootstrapless_synthesizer = BootstraplessSynthesizer()
    stack = Stack(
        app,
        service_name,
        synthesizer=bootstrapless_synthesizer,
        env=Environment(account=aws_account_id, region=aws_region),
    )

    vpc = Vpc.from_vpc_attributes(stack, "vpc", vpc_id=vpc_id, availability_zones=azs)

    subnets = [
        Subnet.from_subnet_id(stack, id=f"container_{i}", subnet_id=subnet_id)
        for i, subnet_id in enumerate(subnet_ids)
    ]

    subnet_selection = SubnetSelection(subnets=subnets)
    task_role = create_task_role(stack, service_name, ecs_manifest)
    execution_task_role = create_execution_task_role(stack, service_name, ecs_manifest)
    ecs_cluster = create_ecs_cluster(stack, service_name, vpc)
    log_group = create_log_group(stack, service_name)
    sg = create_security_group(stack, service_name, vpc)
    task_definition = create_task_definition(
        stack,
        service_name,
        task_role,
        execution_task_role,
        log_group,
        ecs_manifest,
        run,
    )
    create_ecs_service(
        stack, service_name, ecs_cluster, task_definition, subnet_selection, sg
    )

    app.synth()


def create_task_role(stack, service_name, ecs_manifest):
    from aws_cdk.aws_iam import (
        Effect,
        ManagedPolicy,
        PolicyDocument,
        PolicyStatement,
        ServicePrincipal,
        Role,
    )

    role_name = f"{service_name}-task-role"
    policy_name = f"{service_name}-task-role-policy"
    assumed_by = ServicePrincipal("ecs-tasks.amazonaws.com")
    managed_policies_raw = ecs_manifest.role.managed_policies
    managed_policies = [
        ManagedPolicy.from_aws_managed_policy_name(managed_policy)
        for managed_policy in managed_policies_raw
    ]
    policy_statements_raw = ecs_manifest.role.statements
    policy_statements = [
        PolicyStatement(
            sid=policy_statement.sid,
            actions=policy_statement.actions,
            resources=policy_statement.resources,
            effect=(
                Effect.ALLOW if policy_statement.effect == "Allow" else Effect.DENY
            ),
        )
        for policy_statement in policy_statements_raw
    ]
    inline_policies = PolicyDocument(statements=policy_statements)
    role: Role = Role(
        stack,
        role_name,
        assumed_by=assumed_by,
        inline_policies={policy_name: inline_policies},
        managed_policies=managed_policies,
    )
    role.without_policy_updates()
    return role


def create_execution_task_role(stack, service_name, ecs_manifest):
    from aws_cdk.aws_iam import (
        Effect,
        ManagedPolicy,
        PolicyDocument,
        PolicyStatement,
        ServicePrincipal,
        Role,
    )

    role_name = f"{service_name}-execution-task-role"
    policy_name = f"{service_name}-execution-task-role-policy"
    assumed_by = ServicePrincipal("ecs-tasks.amazonaws.com")
    managed_policies_raw = ecs_manifest.execution_role.managed_policies
    managed_policies = [
        ManagedPolicy.from_aws_managed_policy_name(managed_policy)
        for managed_policy in managed_policies_raw
    ]
    policy_statements_raw = ecs_manifest.execution_role.statements
    policy_statements = [
        PolicyStatement(
            sid=policy_statement.sid,
            actions=policy_statement.actions,
            resources=policy_statement.resources,
            effect=(
                Effect.ALLOW if policy_statement.effect == "Allow" else Effect.DENY
            ),
        )
        for policy_statement in policy_statements_raw
    ]
    inline_policies = PolicyDocument(statements=policy_statements)
    role: Role = Role(
        stack,
        role_name,
        assumed_by=assumed_by,
        inline_policies={policy_name: inline_policies},
        managed_policies=managed_policies,
    )
    role.without_policy_updates()
    return role


def create_ecs_cluster(stack, service_name, vpc):
    from aws_cdk.aws_ecs import Cluster

    cluster_name = f"{service_name}-cluster"
    return Cluster(stack, cluster_name, cluster_name=cluster_name, vpc=vpc)


def create_log_group(stack, service_name):
    from aws_cdk.aws_logs import LogGroup

    log_group_name = f"{service_name}-log"
    return LogGroup(stack, log_group_name)


def create_security_group(stack, service_name, vpc):
    from aws_cdk.aws_ec2 import ISecurityGroup, SecurityGroup

    sg_name = f"{service_name}-sg"
    sg: ISecurityGroup = SecurityGroup(stack, sg_name, vpc=vpc, allow_all_outbound=True)
    return sg


def create_ecs_service(
    stack, service_name, cluster, task_definition, subnets, security_group
):
    from aws_cdk.aws_ecs import FargateService

    service_name = f"{service_name}-service"
    return FargateService(
        stack,
        service_name,
        service_name=service_name,
        cluster=cluster,
        task_definition=task_definition,
        vpc_subnets=subnets,
        security_groups=[security_group],
        enable_execute_command=True,
        min_healthy_percent=0,
    )



from pyfzf import FzfPrompt


def select_action(choices, header, multi=False):
    if not multi:
        try:
            selected_action = FzfPrompt().prompt(
                choices,
                fzf_options=f'--height=10% --border=sharp --header="{header}"',
            )[0]
        except IndexError:
            return None
    else:
        selected_action = FzfPrompt().prompt(
            choices,
            fzf_options=f'--height=10% --border=sharp --multi --header="{header}"',
        )
    return selected_action



import json
import os
import random
import time
from typing import Dict
import boto3
import socket
from botocore.exceptions import UnauthorizedSSOTokenError

from easyecs.helpers.color import Color


def load_template(stack_name: str) -> Dict:
    """
    Loads the CloudFormation template from a JSON file.
    """
    with open(f".cloudformation/{stack_name}.template.json") as f:
        return json.load(f)


def parse_dict_with_env_var(build_args):
    for key, value in build_args.items():
        if value.startswith("{{.") and value.endswith("}}"):
            env_var_name = value[3:][:-2]
            new_value = os.environ.get(env_var_name)
            if new_value is None:
                print(
                    f"{Color.RED}Environment variable {env_var_name} does not"
                    f" exist{Color.END}"
                )
            build_args[key] = new_value
    return build_args


def convert_tags_to_dict(tags):
    return {tag["Key"]: tag["Value"] for tag in tags}


def convert_containers_to_dict(containers):
    parsed_containers = {}
    for container in containers:
        container_name = container["name"]
        # We give him a target id so it's easier to use it with ssm start session
        runtime_id = container["runtimeId"]
        task_id = runtime_id.split("-")[0]
        target_id = f"{task_id}_{runtime_id}"
        container["targetId"] = target_id
        parsed_containers[container_name] = container
    return parsed_containers


def is_port_in_use(port: int) -> bool:
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect(("127.0.0.1", int(port)))
        s.shutdown(2)
        return True
    except Exception:
        return False


def check_credentials() -> None:
    client = boto3.client("s3")
    try:
        client.list_buckets()
    except UnauthorizedSSOTokenError:
        print(
            f"{Color.RED}You should refresh your credentials or connect to your aws"
            f" account (aws sso login){Color.END}"
        )
        exit(-1)


def generate_random_port():
    random_port = random.randint(1024, 65535)
    while is_port_in_use(random_port):
        random_port = random.randint(1024, 65535)
        time.sleep(0.1)
    random_port = str(random_port)
    return random_port



#!python

import time
import click
from easyecs.cloudformation.stack.create import create_stack
from easyecs.cloudformation.stack.delete import delete_stack
from easyecs.cloudformation.stack.update import update_stack
from easyecs.cloudformation.template import create_template
from easyecs.cloudformation.fetch import (
    fetch_aws_account,
    fetch_containers,
    fetch_is_stack_created,
)
from easyecs.command import (
    run_nc_commands,
    create_port_forwards,
    execute_command,
    popen_procs_port_forward,
    popen_procs_exec_command,
    threads,
)
from easyecs.docker import build_docker_image
from easyecs.helpers.color import Color
from easyecs.helpers.common import check_credentials
from easyecs.helpers.loader import Loader

from easyecs.helpers.settings import (
    compute_hash_ecs_file,
    delete_hash,
    load_settings,
    read_ecs_file,
    save_hash,
)


def step_import_aws_cdk():
    loader_import = Loader(
        "Importing CloudFormation:",
        "Importing CloudFormation: \u2705",
        "Importing CloudFormation: \u274c",
        0.05,
    )
    loader_import.start()
    # It takes time to import CDK so we show it to the user!
    import aws_cdk  # noqa: F401

    loader_import.stop()


def step_docker_build_and_push(
    no_docker_build,
    ecs_manifest,
    stack_name,
    aws_account_id,
    aws_region,
    vpc_id,
    subnet_ids,
    azs,
    show_docker_logs,
    run,
):
    if not no_docker_build:
        loader_docker = Loader(
            "Building and pushing docker images:",
            "Building and pushing docker images: \u2705",
            "Building and pushing docker images: \u274c",
            0.05,
        )
        loader_docker.start()
        build_docker_image(ecs_manifest, show_docker_logs)
        loader_docker.stop()

    loader = Loader(
        "Creating CloudFormation template:",
        "Creating CloudFormation template: \u2705",
        "Creating CloudFormation template: \u274c",
        0.05,
    )
    loader.start()
    create_template(
        stack_name,
        aws_account_id,
        aws_region,
        vpc_id,
        subnet_ids,
        azs,
        ecs_manifest,
        run,
    )
    loader.stop()


def step_create_or_update_stack(stack_name, force_redeployment):
    if not fetch_is_stack_created(stack_name):
        create_stack(stack_name)
    else:
        update_stack(stack_name, force_redeployment)


def step_idle_keyboard():
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Quitting...")
        pass


def step_clean_exit():
    for popen_proc in popen_procs_port_forward:
        popen_proc.kill()

    for thread in threads:
        thread.stop()
        thread.join()

    for popen_proc in popen_procs_exec_command:
        popen_proc.stdin.write("exit\x03\x04".encode("utf8"))
        popen_proc.stdin.flush()
        popen_proc.wait()


def has_ecs_file_changed(cache_settings):
    hash_sha256 = compute_hash_ecs_file()
    return hash_sha256 != cache_settings["sha256"]


def step_bring_up_stack(
    cache_settings,
    no_docker_build,
    ecs_manifest,
    stack_name,
    aws_account_id,
    aws_region,
    vpc_id,
    subnet_ids,
    azs,
    force_redeployment,
    aws_account,
    show_docker_logs,
    run,
):
    print()
    if has_ecs_file_changed(cache_settings) or force_redeployment:
        step_import_aws_cdk()
        step_docker_build_and_push(
            no_docker_build,
            ecs_manifest,
            stack_name,
            aws_account_id,
            aws_region,
            vpc_id,
            subnet_ids,
            azs,
            show_docker_logs,
            run,
        )
        step_create_or_update_stack(stack_name, force_redeployment)
        save_hash(aws_account)
    else:
        print(f"{Color.YELLOW}No updates are to be performed.{Color.END}")


def action_run(ctx):
    no_docker_build = ctx.obj["no_docker_build"]
    force_redeployment = ctx.obj["force_redeployment"]
    show_docker_logs = ctx.obj["show_docker_logs"]
    aws_account = fetch_aws_account()
    cache_settings = load_settings(aws_account)
    ecs_manifest = read_ecs_file()
    app_name = ecs_manifest.metadata.appname
    user = ecs_manifest.metadata.user
    aws_region = cache_settings["aws_region"]
    aws_account_id = cache_settings["aws_account_id"]
    vpc_id = cache_settings["vpc_id"]
    subnet_ids = cache_settings["subnet_ids"]
    azs = cache_settings["azs"]
    stack_name = f"{user}-{app_name}"

    step_bring_up_stack(
        cache_settings,
        no_docker_build,
        ecs_manifest,
        stack_name,
        aws_account_id,
        aws_region,
        vpc_id,
        subnet_ids,
        azs,
        force_redeployment,
        aws_account,
        show_docker_logs,
        run=True,
    )
    parsed_containers = fetch_containers(user, app_name)
    print()
    create_port_forwards(ecs_manifest, aws_region, aws_account, parsed_containers)
    step_idle_keyboard()

    step_clean_exit()

    exit(0)


def action_dev(ctx):
    no_docker_build = ctx.obj["no_docker_build"]
    force_redeployment = ctx.obj["force_redeployment"]
    show_docker_logs = ctx.obj["show_docker_logs"]
    aws_account = fetch_aws_account()
    cache_settings = load_settings(aws_account)
    ecs_manifest = read_ecs_file()
    app_name = ecs_manifest.metadata.appname
    user = ecs_manifest.metadata.user
    aws_region = cache_settings["aws_region"]
    aws_account_id = cache_settings["aws_account_id"]
    vpc_id = cache_settings["vpc_id"]
    subnet_ids = cache_settings["subnet_ids"]
    azs = cache_settings["azs"]
    stack_name = f"{user}-{app_name}"

    step_bring_up_stack(
        cache_settings,
        no_docker_build,
        ecs_manifest,
        stack_name,
        aws_account_id,
        aws_region,
        vpc_id,
        subnet_ids,
        azs,
        force_redeployment,
        aws_account,
        show_docker_logs,
        run=False,
    )
    parsed_containers = fetch_containers(user, app_name)
    print()
    create_port_forwards(ecs_manifest, aws_region, aws_account, parsed_containers)
    run_nc_commands(parsed_containers, aws_region, aws_account, ecs_manifest)
    print()

    found_tty = execute_command(
        ecs_manifest,
        parsed_containers,
        aws_region,
        aws_account,
    )

    if not found_tty:
        step_idle_keyboard()
    step_clean_exit()
    exit(0)


def action_delete(_):
    aws_account = fetch_aws_account()
    ecs_manifest = read_ecs_file()
    app_name = ecs_manifest.metadata.appname
    user = ecs_manifest.metadata.user
    stack_name = f"{user}-{app_name}"
    delete_stack(stack_name)
    delete_hash(aws_account)


@click.group()
@click.pass_context
def entrypoint(ctx):
    ctx.ensure_object(dict)
    check_credentials()


@entrypoint.command(name="run", help="Run a stack")
@click.option(
    "--no-docker-build",
    is_flag=True,
    default=False,
    show_default=True,
    help=(
        "If used, easyecs will not build and push docker image of containers if there"
        " is one to build."
    ),
)
@click.option(
    "--force-redeployment",
    is_flag=True,
    default=False,
    show_default=True,
    help=(
        "If used, and only when there's no update on the cloudformation stack, easyecs"
        " will force a new deployment of the task."
    ),
)
@click.option(
    "--show-docker-logs",
    is_flag=True,
    default=False,
    show_default=True,
    help="If used, it will show the docker build and push logs",
)
@click.pass_context
def click_run(ctx, no_docker_build, force_redeployment, show_docker_logs):
    ctx.obj["no_docker_build"] = no_docker_build
    ctx.obj["force_redeployment"] = force_redeployment
    ctx.obj["show_docker_logs"] = show_docker_logs
    action_run(ctx)


@entrypoint.command(name="dev", help="Run a stack in development mode")
@click.option(
    "--no-docker-build",
    is_flag=True,
    default=False,
    show_default=True,
    help=(
        "If used, easyecs will not build and push docker image of containers if there"
        " is one to build."
    ),
)
@click.option(
    "--force-redeployment",
    is_flag=True,
    default=False,
    show_default=True,
    help=(
        "If used, and only when there's no update on the cloudformation stack, easyecs"
        " will force a new deployment of the task."
    ),
)
@click.option(
    "--show-docker-logs",
    is_flag=True,
    default=False,
    show_default=True,
    help="If used, it will show the docker build and push logs",
)
@click.pass_context
def click_dev(ctx, no_docker_build, force_redeployment, show_docker_logs):
    ctx.obj["no_docker_build"] = no_docker_build
    ctx.obj["force_redeployment"] = force_redeployment
    ctx.obj["show_docker_logs"] = show_docker_logs
    action_dev(ctx)


@entrypoint.command(name="delete", help="Delete a stack")
@click.pass_context
def click_delete(ctx):
    action_delete(ctx)


if __name__ == "__main__":
    entrypoint()



import json
from typing import Dict
import boto3
from botocore.utils import ClientError
from easyecs.cloudformation.client import get_client_cloudformation
from easyecs.cloudformation.fetch import fetch_stack_url
from easyecs.cloudformation.stack.waiter import (
    wait_for_stack_create,
    wait_for_stack_rollback,
    wait_for_stack_update,
)
from easyecs.command import run_force_new_deployment

from easyecs.helpers.color import Color
from easyecs.helpers.common import load_template
from easyecs.helpers.loader import Loader


def update_cloudformation_stack(stack_name: str, template_body: Dict):
    """
    Sends a request to AWS to update a CloudFormation stack.
    """
    client = get_client_cloudformation()
    client.update_stack(
        StackName=stack_name,
        TemplateBody=json.dumps(template_body),
        Capabilities=["CAPABILITY_NAMED_IAM"],
    )


def handle_update_error(
    e: ClientError, stack_name: str, force_redeployment: bool, loader: Loader
):
    """
    Handles a CloudFormation stack update failure.
    Depending on the error, either waits for a rollback, cancels an update,
    or prints the error and breaks the loop.
    """
    message = e.response["Error"]["Message"]
    if message == "No updates are to be performed.":
        loader.stop()
        print(f"{Color.YELLOW}No updates are to be performed.{Color.END}")
        if force_redeployment:
            run_force_new_deployment(stack_name)
    elif "UPDATE_IN_PROGRESS" in message:
        cloudformation = boto3.resource("cloudformation")
        stack = cloudformation.Stack(stack_name)
        stack.cancel_update()
        wait_for_stack_rollback(stack_name)
        loader.stop()
    elif "ROLLBACK_IN_PROGRESS" in message:
        wait_for_stack_rollback(stack_name)
        loader.stop()
    elif "CREATE_IN_PROGRESS" in message:
        wait_for_stack_create(stack_name)
        loader.stop()
    else:
        loader.stop_error()
        print(e)


def update_stack(stack_name: str, force_redeployment: bool):
    """
    Updates a CloudFormation stack with the given name.
    """
    loader = Loader(
        "Updating CloudFormation stack:",
        "Updating CloudFormation stack: \u2705",
        "Updating CloudFormation stack: \u274c",
        0.05,
    )
    loader.start()

    cloudformation_template = load_template(stack_name)
    loader.set_metadata(f"Cloudformation URL: {fetch_stack_url(stack_name)}")

    try:
        update_cloudformation_stack(stack_name, cloudformation_template)
        wait_for_stack_update(stack_name)
        loader.stop()
    except ClientError as e:
        handle_update_error(e, stack_name, force_redeployment, loader)



import hashlib
import json
import os
import yaml
from yaml import SafeLoader

from easyecs.cloudformation.fetch import (
    fetch_account_id,
    fetch_availability_zones,
    fetch_container_subnet_ids,
    fetch_region,
    fetch_vpc_id,
)
from easyecs.helpers.color import Color
from easyecs.model.ecs import EcsFileModel


def read_ecs_file() -> EcsFileModel:
    with open("./ecs.yml") as f:
        data = yaml.load(f, Loader=SafeLoader)
    return EcsFileModel(**data)


def compute_hash_ecs_file():
    hash_sha256 = hashlib.sha256()
    with open("ecs.yml", "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


def save_hash(aws_account):
    hash_sha256 = compute_hash_ecs_file()
    with open(".tmp/easyecs.tmp", "r") as f:
        cache_configuration = json.load(f)
    cache_configuration[aws_account]["sha256"] = hash_sha256
    with open(".tmp/easyecs.tmp", "w") as f:
        json.dump(cache_configuration, f)


def delete_hash(aws_account):
    with open(".tmp/easyecs.tmp", "r") as f:
        cache_configuration = json.load(f)
    cache_configuration[aws_account]["sha256"] = None
    with open(".tmp/easyecs.tmp", "w") as f:
        json.dump(cache_configuration, f)


def load_settings(aws_account):
    try:
        os.makedirs(".tmp", exist_ok=True)
        with open(".tmp/easyecs.tmp", "r") as f:
            cache_configuration = json.load(f)
    except FileNotFoundError:
        cache_configuration = {}

    account_cache_configuration = cache_configuration.get(aws_account, {})

    aws_region = account_cache_configuration.get("aws_region", None)
    if not aws_region:
        aws_region = fetch_region()
    azs = account_cache_configuration.get("azs", None)
    if not azs:
        azs = fetch_availability_zones(aws_region)
    aws_account_id = account_cache_configuration.get("aws_account_id", None)
    if not aws_account_id:
        aws_account_id = fetch_account_id()
    vpc_id = account_cache_configuration.get("vpc_id", None)
    if not vpc_id:
        vpc_id = fetch_vpc_id()
    subnet_ids = account_cache_configuration.get("subnet_ids", None)
    if not subnet_ids:
        subnet_ids = fetch_container_subnet_ids(vpc_id)

    hash_sha256 = account_cache_configuration.get("sha256", None)

    cache_configuration[aws_account] = {
        "aws_region": aws_region,
        "aws_account_id": aws_account_id,
        "vpc_id": vpc_id,
        "subnet_ids": subnet_ids,
        "azs": azs,
        "sha256": hash_sha256,
    }

    print(
        "Selected aws account:"
        f" {Color.BOLD}{Color.GREEN}{aws_account}{Color.END} \u2705"
    )
    print(f"Selected region: {Color.BOLD}{Color.GREEN}{aws_region}{Color.END} \u2705")
    print(f"Selected vpc_id: {Color.BOLD}{Color.GREEN}{vpc_id}{Color.END} \u2705")
    print(
        "Selected subnet_ids:"
        f" {Color.BOLD}{Color.GREEN}{','.join(subnet_ids)}{Color.END} \u2705"
    )
    print(
        "Selected availability zones:"
        f" {Color.BOLD}{Color.GREEN}{','.join(azs)}{Color.END} \u2705"
    )

    os.makedirs(".tmp", exist_ok=True)
    with open(".tmp/easyecs.tmp", "w") as f:
        json.dump(cache_configuration, f)

    return cache_configuration[aws_account]



import os
from typing import Dict, List, Optional, Union
from pydantic import BaseModel, computed_field, field_validator


class EcsFileMetadataModel(BaseModel):
    appname: str
    user: str = os.environ["USER"]


class EcsFileStatementModel(BaseModel):
    sid: str
    resources: List[str]
    actions: List[str]
    effect: str


class EcsFileRoleModel(BaseModel):
    managed_policies: List[str] = []
    statements: List[EcsFileStatementModel]

    @field_validator("statements")
    def validate_unique_sid(cls, statements):
        seen = set()
        for statement in statements:
            if statement.sid in seen:
                raise ValueError(f"Duplicate sid {statement.sid} found in statements.")
            seen.add(statement.sid)
        return statements


class EcsFileLimitsModel(BaseModel):
    cpu: int
    memory: int


class EcsFileResourcesModel(BaseModel):
    limits: EcsFileLimitsModel


class EcsFileBuildModel(BaseModel):
    dockerfile: str = "Dockerfile"
    target: Optional[str] = None
    args: Dict[str, str] = {}


class EcsFileEnvModel(BaseModel):
    name: str
    value: str
    active: bool = True


class EcsFileSecretModel(BaseModel):
    name: str
    arn: str
    field: str
    active: bool = True


class EcsFileVolumeModel(BaseModel):
    name: str
    id: str
    mount_point: str


class EcsFileContainerHealthCheckModel(BaseModel):
    command: Union[List[str], str]
    interval: int
    retries: int
    start_period: int
    timeout: int


class EcsFileContainerModel(BaseModel):
    name: str
    image: str
    user: str = "root"
    tty: bool = False
    essential: bool = True
    entry_point: Union[Optional[str], Optional[List[str]]] = None
    command: Union[Optional[str], Optional[List[str]]] = None
    resources: EcsFileResourcesModel
    build: Optional[EcsFileBuildModel] = None
    port_forward: List[str] = []
    env: List[EcsFileEnvModel] = []
    secrets: List[EcsFileSecretModel] = []
    efs_volumes: List[EcsFileVolumeModel] = []
    volumes: List[str] = []
    healthcheck: Optional[EcsFileContainerHealthCheckModel] = None
    depends_on: Optional[Dict[str, Dict[str, str]]] = None


class EcsTaskDefinitionModel(BaseModel):
    resources: EcsFileResourcesModel
    containers: List[EcsFileContainerModel]

    @field_validator("containers")
    def validate_single_tty(cls, containers):
        tty_containers = [container.name for container in containers if container.tty]
        if len(tty_containers) > 1:
            raise ValueError(
                "More than one container has tty set to true:"
                f" {', '.join(tty_containers)}."
            )
        return containers

    @computed_field(return_type=List[EcsFileVolumeModel])
    @property
    def efs_volumes(self) -> List[EcsFileVolumeModel]:
        efs_volumes = [
            volume for container in self.containers for volume in container.efs_volumes
        ]
        return efs_volumes


class EcsFileModel(BaseModel):
    metadata: EcsFileMetadataModel
    role: EcsFileRoleModel
    execution_role: EcsFileRoleModel
    task_definition: EcsTaskDefinitionModel



import json
from typing import Dict

from botocore.waiter import WaiterError
from easyecs.cloudformation.client import get_client_cloudformation
from easyecs.cloudformation.fetch import fetch_stack_url
from easyecs.cloudformation.stack.delete import delete_stack

from easyecs.helpers.color import Color
from easyecs.helpers.common import load_template
from easyecs.helpers.loader import Loader


def create_cloudformation_stack(stack_name: str, template_body: Dict):
    """
    Sends a request to AWS to create a CloudFormation stack.
    """
    client = get_client_cloudformation()
    client.create_stack(
        StackName=stack_name,
        TemplateBody=json.dumps(template_body),
        Capabilities=["CAPABILITY_NAMED_IAM"],
    )


def wait_for_stack_creation(stack_name: str):
    """
    Waits for the CloudFormation stack to be created.
    Throws an exception if the creation fails.
    """
    client = get_client_cloudformation()
    waiter = client.get_waiter("stack_create_complete")
    waiter.wait(StackName=stack_name)


def handle_stack_creation_failure(e: WaiterError, stack_name: str):
    """
    Handles a CloudFormation stack creation failure.
    Prints the reason for the failure, deletes the stack, and exits the program.
    """
    reason = e.kwargs["reason"]
    print(f"{Color.RED}{reason}{Color.END}")
    delete_stack(stack_name)
    exit(-1)


def create_stack(stack_name: str):
    """
    Creates a CloudFormation stack with the given name.
    """
    loader = Loader(
        "Creating CloudFormation stack:",
        "Creating CloudFormation stack: \u2705",
        "Creating CloudFormation stack: \u274c",
        0.05,
    )
    loader.start()

    try:
        cloudformation_template = load_template(stack_name)
        create_cloudformation_stack(stack_name, cloudformation_template)
        loader.set_metadata(f"Cloudformation URL: {fetch_stack_url(stack_name)}")
        wait_for_stack_creation(stack_name)
    except WaiterError as e:
        handle_stack_creation_failure(e, stack_name)

    loader.stop()



import subprocess
from easyecs.helpers.common import parse_dict_with_env_var


def docker_build_cmd(build, image_name):
    dockerfile = build.dockerfile
    target = build.target
    if target:
        target = f"--target {target}"
    build_args = build.args
    build_args = parse_dict_with_env_var(build_args)
    build_args_str = " ".join(
        [f'--build-arg {key}="{value}"' for key, value in build_args.items()]
    )
    build_cmd = f"docker build -t {image_name}"
    if dockerfile:
        build_cmd += f" -f {dockerfile}"
    if target:
        build_cmd += f" {target}"
    if build.args:
        build_cmd += f" {build_args_str}"
    build_cmd += " ."
    return build_cmd


def build_docker_image(ecs_manifest, show_docker_logs):
    containers = ecs_manifest.task_definition.containers
    for container in containers:
        image_name = container.image
        build = container.build
        if build:
            build_cmd = docker_build_cmd(build, image_name)
            if show_docker_logs:
                res = subprocess.Popen(
                    build_cmd,
                    shell=True,
                )
            else:
                res = subprocess.Popen(
                    build_cmd,
                    shell=True,
                    stderr=subprocess.DEVNULL,
                    stdout=subprocess.DEVNULL,
                )
            res.wait()
            if res.poll() != 0:
                raise Exception(
                    "There was an issue building the docker image. Use"
                    " --show-docker-logs to get more information!"
                )
            push_docker_image(image_name, show_docker_logs)


def push_docker_image(image_name, show_docker_logs):
    push_cmd = f"docker push {image_name}"
    if show_docker_logs:
        res = subprocess.Popen(push_cmd, shell=True)
    else:
        res = subprocess.Popen(
            push_cmd, shell=True, stderr=subprocess.DEVNULL, stdout=subprocess.DEVNULL
        )
    res.wait()
    if res.poll() != 0:
        raise Exception(
            "There was an issue pushing the docker image. Use --show-docker-logs to get"
            " more information!"
        )



import datetime
from os.path import dirname
import subprocess
from watchdog.events import FileSystemEvent, FileSystemEventHandler


class SynchronizeEventHandler(FileSystemEventHandler):
    def __init__(self, volume, port):
        super().__init__()
        self.volume = volume
        self.port = port
        self.input = volume.split(":")[0]
        self.output = volume.split(":")[1]
        self.input_dirname = dirname(self.input)
        self.output_dirname = dirname(self.output)
        self.last_event = datetime.datetime.now().timestamp()

    def synchronize(self):
        cmd_nc_local = ["nc", "-N", "127.0.0.1", self.port]
        if self.input_dirname.startswith("/"):
            cmd_input_dirname = self.input_dirname[1:]
        else:
            cmd_input_dirname = self.input_dirname
        tar_cmd = [
            "tar",
            "-czvf",
            "-",
            self.input,
            f"--transform=s,{cmd_input_dirname}/,{self.output_dirname}/,",
        ]
        proc_tar_local = subprocess.run(
            tar_cmd,
            start_new_session=True,
            check=True,
            capture_output=True,
        )
        subprocess.run(
            cmd_nc_local,
            start_new_session=True,
            input=proc_tar_local.stdout,
            stdout=subprocess.DEVNULL,
        )

    def dispatch(self, event: FileSystemEvent):
        delta = datetime.datetime.now().timestamp() - self.last_event
        if delta >= 1:
            event_src = event.src_path
            if event.event_type in ["modified", "created", "moved", "deleted"]:
                if event.is_directory:
                    if self.input in event_src:
                        self.synchronize()
                        super().dispatch(event)
                        self.last_event = datetime.datetime.now().timestamp()
                else:
                    if event_src in self.input:
                        self.synchronize()
                        super().dispatch(event)
                        self.last_event = datetime.datetime.now().timestamp()
